/*

   Copyright 2010, 2011 SRI International

   Licensed under the Apache License, Version 2.0 (the "License"); you may
   not use this file except in compliance with the License. You may obtain
   a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   The development was supported by National Science Foundation Grant 0932397:
   A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package lf.api;

import java.io.FileInputStream;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberNode;

import alice.tuprolog.ClauseInfo;
import alice.tuprolog.Prolog;
import alice.tuprolog.SolveInfo;
import alice.tuprolog.Struct;
import alice.tuprolog.Term;
import alice.tuprolog.Theory;
import alice.tuprolog.TheoryManager;

public class CPSLogic extends CyberApp implements CPSLogicEngine 
{
	Prolog logic_engine; 
	Theory ForwardRules;
	Theory BackwardRules; 

	boolean RandomGoalSelection;
    boolean noAlternatives;

    String ForwardRuleFile;
    String BackwardRuleFile;

	GoalFactory cpsGoalFactory;
	FactFactory cpsFactFactory;

	public CPSLogic(CyberNode node, boolean r, boolean a, String fr, String br) 
 	{
		super(node); 

		this.RandomGoalSelection = r;
    	this.noAlternatives = a;

    	this.ForwardRuleFile = fr;
    	this.BackwardRuleFile = br;

		try {
			logic_engine = new Prolog(new String[]{"alice.tuprolog.lib.BasicLibrary", "alice.tuprolog.lib.CPSLibrary"});

            ForwardRules = new Theory(new FileInputStream(ForwardRuleFile));
            BackwardRules = new Theory(new FileInputStream(BackwardRuleFile));
            logic_engine.setTheory(ForwardRules, "forwardRules");
            logic_engine.addTheory(BackwardRules, "backwardRules");
        } catch(Exception e) {
            System.out.println("Exception in LogicApp constructor");
            System.exit(0);
		}
	}

	public void setFactory(GoalFactory goalfactory, FactFactory factfactory)
	{
		cpsGoalFactory = goalfactory;
		cpsFactFactory = factfactory;
	}

	public void theoryUpdate(AppKnowledgeBase akb)
    {
        try {
            logic_engine.setTheory(ForwardRules, "forwardRules");
            logic_engine.addTheory(BackwardRules, "backwardRules");

            ListIterator<ApplicationKnowledge> li = akb.listIterator();
            while(li.hasNext())
            {
                ApplicationKnowledge k = li.next();
                /* if(k instanceof Goal)
                {
                    Goal g = (Goal) k;

                    //log(name() + " - theoryUpdate goal " + g.lf());
                }
				*/
                if(k instanceof Fact)
                {
                    Fact f = (Fact) k;

                    //log(name() + " - theoryUpdate fact " + f.term());

                    Theory new_th = new Theory(f.term());
                    logic_engine.addTheory(new_th, "fact");
                }
            }
        } catch(Exception x) {
            log(name() + " - Exception in theoryUpdate");
            System.exit(0);
        }
    }

    public String selectTheoryGoal(AppKnowledgeBase akb)
    {
		boolean randomized = RandomGoalSelection; 
		boolean dynamic = false; 

    	int cnt = 0, s_cnt = 0;
    	ListIterator<ApplicationKnowledge> li = akb.listIterator();
    	while(li.hasNext())
    	{
        	ApplicationKnowledge k = li.next();
        	if(k instanceof Goal)
        	{
            	Goal g = (Goal) k;

            	if(!dynamic)
            	{
                	if(!randomized) return g.lf();
                	cnt++;
            	}
            	else
            	{
                	/* unsatisfied goal is favored */
                	if(!g.satisfied())
                	{
                    	if(!randomized) return g.lf();
                    	cnt++;
                	}
            	}
        	}
    	}

    	if(cnt > 0)
    	{
            int n = (int)((double)cnt * Math.random());
        	cnt = 0;
        	li = akb.listIterator();
        	while(li.hasNext())
        	{
            	ApplicationKnowledge k = li.next();
            	if(k instanceof Goal)
            	{
                	Goal g = (Goal) k;

                	if(!dynamic)
                	{
                    	if(n <= cnt++) return g.lf();
                	}
                	else
                	{
                    	/* unsatisfied goal is favored */
                    	if(!g.satisfied())
                        	if(n <= cnt++) return g.lf();
                	}
            	}
        	}
    	}

    	return null;
    }

	public void reason(Goal g)
    {
        try{
            //log(name() + " - theory try to solve " + g.lf());
            //log(name() + " - theory\n" + logic_engine.getTheory().toString());

            //EngineManager egm = logic_engine.getEngineManager();
            //log(name() + " - context\n" + egm.getCurrentContext().toString());

            SolveInfo info = logic_engine.solve(g.lf());
            while(info.isSuccess())
            {
                //log(name() + " - " + info.getSolution() + " satisfied");
                //log(name() + " - bindings: " + info);

                // add a new fact 
                String tr;
                if(info.getBindingSize() > 0) {
                    tr = info.getSolution().toString() + ".";
                }
                else {
                    tr = g.lf();
                }
				FactType ftype = getFactType(tr);
                postKnowledge(cpsFactFactory.getFact(this, tr, ftype));

                // mark a goal is satisfied 
                // Goal newg = new Goal(this,g.lf(),true);
                // postKnowledge(newg);
                /*
                Iterator it = info.getBindingVars().iterator();
                while (it.hasNext()) {
                   	Var v = (Var)it.next();
                	log(name() + " - binding " + info.getVarValue(v.getName()).toString());
                }
                */
                if(noAlternatives)
                    break;

                // check alternative solutions 
                if (logic_engine.hasOpenAlternatives()) {
                    //log(name() + " - alternative binding exists for " + g.lf());
                    info = logic_engine.solveNext();
                } else {
                    //log(name() + " - no alternative binding exists for " + g.lf());
                    break;
                }
            }

            if (info.getFailedSubGoal() != null)
            {
                String sg = info.getFailedSubGoal().toString() + ".";
                //log(name() + " - failed subgoal " + sg);
				if(!isForwardRule(sg))
                	backwardSubgoalPropagation(sg);
            }
            //backwardReasoning (info);
            forwardReasoning (true);
       	} catch (Exception x) {x.printStackTrace(); System.exit(0);}
    }

	public void backwardSubgoalPropagation(String lf)
    {
    	//if (from staticDB) no need to put in the KB   
    	TheoryManager thm = logic_engine.getTheoryManager();
    	Term headt = Term.createTerm(lf.substring(0,lf.length()-1));
    	List list = thm.findStaticDB(headt);
    	Struct st = (Struct) headt;
    	String s = st.getName();
    	if (!list.isEmpty() && !s.equals("move"))
			;
        	//log(name() + " - basic op " + list);
  		else
		{
			GoalType type = getGoalType(lf);
    		postKnowledge(cpsGoalFactory.getGoal(this, lf, type));
		}
    }

    void forwardReasoning (Boolean randomized)
    {
    	Struct c = selectRule(randomized);
		if (c != null) 
    	{
			String lf = c.getArg(0).toString() + ".";
    		// log(name() + " - forward reasoing new subgoal: " + lf);

			try {
				SolveInfo info = logic_engine.solve(lf);
            	while(info.isSuccess())
            	{
                	//log(name() + " - " + info.getSolution() + " satisfied");
                	//log(name() + " - bindings: " + info);

                	// add a new fact 
                	String tr;
                	if(info.getBindingSize() > 0) {
                    	tr = info.getSolution().toString() + ".";
                	}
                	else {
                    	tr = lf;
                	}
                	FactType ftype = getFactType(tr);
                	postKnowledge(cpsFactFactory.getFact(this, tr, ftype));

                	if(noAlternatives)
                    	break;

                	// check alternative solutions 
                	if (logic_engine.hasOpenAlternatives()) {
                    	//log(name() + " - alternative binding exists for " +
                    	//g.lf());
                    	info = logic_engine.solveNext();
                	} else {
                    	//log(name() + " - no alternative binding exists for " +
                    	//g.lf());
                    	break;
                	}
            	}
       		} catch (Exception x) {x.printStackTrace(); System.exit(0);}
		}
    }

    void backwardReasoning (SolveInfo info)
    {
        TheoryManager thm = logic_engine.getTheoryManager();
        for (Iterator it = thm.dynamicDBase.iterator(); it.hasNext();)
    	{
            ClauseInfo d = (ClauseInfo) it.next();
        	//log(name() + " - query : " + info.getQuery() + " " + d.getHead().toString());
        	if(info.getQuery().unify(logic_engine, d.getHead()) == true)
        	{
            	//log(name() + " - unify : " + d.getHead().toString());
                for (Iterator bit = d.getBody().iterator(); bit.hasNext();) {
            		String lf = bit.next().toString() + ".";
					GoalType type = getGoalType(lf);
    				postKnowledge(cpsGoalFactory.getGoal(this, lf, type));
            	}
        	}
    	}
    }

	GoalType getGoalType(String s)
	{
		if (s.startsWith("interest"))		return GoalType.Interest;
		if (s.startsWith("adjacent"))		return GoalType.Adjacent;
		if (s.startsWith("trigger"))		return GoalType.Trigger;
		if (s.startsWith("takesnapshot"))	return GoalType.Takesnapshot;
		if (s.startsWith("snapshot"))		return GoalType.Snapshot;
		if (s.startsWith("result"))			return GoalType.Result;
		if (s.startsWith("deliver"))		return GoalType.Deliver;
		if (s.startsWith("moveto"))			return GoalType.MoveTo; 
		if (s.startsWith("move"))			return GoalType.Move;

		return GoalType.Undefined;
	}	

	FactType getFactType(String s)
	{
		if (s.startsWith("position"))		return FactType.Position;
		if (s.startsWith("adjacent"))		return FactType.Adjacent;
		if (s.startsWith("trigger"))		return FactType.Trigger;
		if (s.startsWith("snapshot"))		return FactType.Snapshot;
		if (s.startsWith("result"))			return FactType.Result;
		if (s.startsWith("deliver"))		return FactType.Deliver;
		if (s.startsWith("noise"))			return FactType.Noise;
		if (s.startsWith("motion"))			return FactType.Motion;

		return FactType.Undefined;
	}	

	boolean isForwardRule(String lf)
	{
    	TheoryManager thm = logic_engine.getTheoryManager();
        for (Iterator it = thm.dynamicDBase.iterator(); it.hasNext();)
    	{
            ClauseInfo d = (ClauseInfo) it.next();
        	Term t = Term.createTerm(lf.substring(0,lf.length()-1));

        	if((d.getLibraryName().equals("forwardRules")) && (t.match(d.getHead()) == true))   
					return true;
		}
		return false;  
	}
 
    Struct selectRule(Boolean randomized)    
	{
    	int cnt = 0;
    	TheoryManager thm = logic_engine.getTheoryManager();
        for (Iterator it = thm.dynamicDBase.iterator(); it.hasNext();)
    	{
            ClauseInfo d = (ClauseInfo) it.next();
        	Struct c = d.getClause();
        	// log(name() + " - clause : " + c.toString() + " libName " + d.getLibraryName());
        	if(!c.getArg(1).toString().equals("true") && d.getLibraryName().equals("forwardRules")) // subgoal 
        	{
            	if(!randomized)
                	return c;
            	cnt++;
        	}
    	}

    	if(cnt > 0)    
		{
        	int n = (int)((double)cnt * Math.random());
            cnt = 0;
            for (Iterator it = thm.dynamicDBase.iterator(); it.hasNext();)
        	{
                ClauseInfo d = (ClauseInfo) it.next();
            	Struct c = d.getClause();
            	//log(name() + " - clause : " + c.toString());
            	if(!c.getArg(1).toString().equals("true") && d.getLibraryName().equals("forwardRules")) // subgoal 
            	{
                	if(n <= cnt++) return c;
            	}
        	}
    	}

    	return null;
    }
}

